/************************ptru************************************************
 *
 *  Frexx Database
 *
 *  A simple, lightweight database 
 *
 ************************************************************************/
#include "dbase.h"
#include "dbase64.h"

/*
** --- major function prototypes ---
*/
int addrow( int argc, char* argv[] );
int printdb( int argc, char* argv[] );
int delrow( int argc, char* argv[] );
int modify( int argc, char* argv[] );
int addfield( int argc, char* argv[] );
int find( int argc, char* argv[] );

/*
** --- global variables ---
*/
LIST dbTemplate;
LIST rowList;

/*
** -------- code ------------
*/
int main( int argc, char* argv[] )
{
    int rc;

    if ( argc==1 ) {
        printf("Frexx Database " VERSION "\n");
        printf("usage: %s [command] [params]\n\n", argv[0] );
        printf("commands: add      [database] {-u} [field id #1] [data #1/@file1] ...\n");
        printf("          find     [database] [search expression] {-u | -h} [format]\n");
        printf("          print    [database] {-i rowid} {-u | -h} [format]\n");
        printf("          delete   [database] [row id]\n");
        printf("          modify   [database] [row id] {-u} [field #1] [data #1/@file #1] ...\n");
        printf("          addfield [database] [name] [type]\n");
        return -1;
    }

    if ( !strcmp( argv[1], "find" ) )
        rc = find( argc, argv );
    else
        if ( !strcmp( argv[1], "print" ) )
            rc = printdb( argc, argv );
        else
            if ( !strcmp( argv[1], "add" ) )
                rc = addrow( argc, argv );
            else
                if ( !strcmp( argv[1], "delete" ) )
                    rc = delrow( argc, argv );
                else
                    if ( !strcmp( argv[1], "modify" ) )
                        rc = modify( argc, argv );
                    else
                        if( !strcmp( argv[1], "addfield" ) )
                            rc = addfield( argc, argv );
                        else {
                            printf("Unknown command '%s'\n", argv[1] );
                            rc = -1;
                        }

    return rc;
}

char* string2html( unsigned char* string )
{
    int i;
    int len=strlen(string);
    unsigned char* url;
    unsigned char* ptr;

    url = (unsigned char*) malloc( len * 6 );
    ptr = url;
    
    for ( i=0;i<len;i++ ) {
        if ( isalnum(string[i])) {
            *ptr=string[i];
            ptr++;
        }
        else {
            sprintf(ptr, "&#%u;", string[i]);
            if ( string[i] < 10 )
                ptr+=4;
            else
                if ( string[i] < 100 )
                    ptr+=5;
                else
                    ptr+=6;
        }
    }
    *ptr = 0;

    return url;
}

char* string2url( char* string )
{
    int i;
    int len=strlen(string);
    char* url;
    char* ptr;

    url = (char*) malloc( len * 3 );
    ptr = url;
    
    for ( i=0;i<len;i++ ) {
        if ( isalnum(string[i])) {
            *ptr=string[i];
            ptr++;
        }
        else {
            if ( ' ' == string[i] ) {
                *ptr = '+';
                ptr++;
            }
            else {
                sprintf(ptr, "%%%02X", string[i]);
                ptr+=3;
            }
        }
    }
    *ptr = 0;

    return url;
}

char* url2string( char* url )
{
    int len,x;
    int val;
    char* ptr;
    char* data;

    len=strlen(url);
    data=(unsigned char*) malloc( len + 1 );
    ptr=data;

    for ( x=0; x<len; x++ ) {
        switch ( url[x] ) {
          case '%':
              sscanf( url+x+1, "%02X", &val );
              *ptr = val;
              ptr++;
              x+=2;
              break;

          case '+':
              *ptr = ' ';
              ptr++;
              break;
              
          default:
              *ptr = url[x];
              ptr++;
        }
    }
    *ptr=0;

    return data;
}

/* check file size */
int filesize( char* filename )
{
    struct stat buf;

    stat( filename, &buf );

    return buf.st_size;
}

/* find a field from its' name */
t_field* getfield( t_row* row, char* name )
{
    t_field* f;

    /* find field */
    for ( f = (t_field*) GetFirst( &row->fieldList ); f;
          f = (t_field*) GetNext( &f->node ) ) {
        
        if ( !strcmp( name, f->template->name ) )
            return f;
    }
    
    return NULL;
}

/* write one row to an open file */
int writerow( FILE* file, char* filename, t_row* row )
{
    t_field* field;

    /* store new-row marker */
    fprintf( file, "- %08lx%08lx\n", row->id1, row->id2 );

    /* store row */
    for ( field = (t_field*) GetFirst( &row->fieldList );
          field;
          field = (t_field*) GetNext( &field->node ) ) {

        switch ( field->template->type ) {
          case integer:
              if ( 0 > fprintf(file, "%s:0:%d\n",
                               field->template->name,
                               field->value ) ) {
                  perror( filename );
                  return -1;
              }
              break;

          case string: 
              {
                  int i,col=0;

				  /* field->len = strlen( field->data ); */

                  if ( 0 > fprintf(file, "%s:%d:\n:",
                                   field->template->name,
                                   field->len ) ) {
                      perror( filename );
                      return -1;
                  }

				  
                  for ( i=0; i<field->len; i++ ) {

                      /* wrap lines */
                      if ( col > 70 ) {
                          if ( 0 > fprintf(file,"\n:") ) {
                              perror( filename );
                              return -1;
                          }
                          col=1;
                      }

                      /* check for backslash */
                      if ( '\\' == field->data[i] ) {
                          if ( 0 > fprintf(file,"\\\\") ) {
                              perror( filename );
                              return -1;
                          }
                          col += 2;
                      }
                      else {
                                /* just print all normal chars */
                          if ( field->data[i] >= 32 ) {
                              if ( 0 > fprintf(file,"%c", field->data[i]) ) {
                                  perror( filename );
                                  return -1;
                              }
                              col++;
                          }

                                /* we must escape control characters */
                          else {
                              int c;
                              switch ( field->data[i] ) {
                              case '\n': c='n'; break;
                              case '\t': c='t'; break;
                              case '\r': c='r'; break;
                              default: c=0; break;
                              }

                              if ( c ) {
                                  if ( 0 > fprintf(file,"\\%c", c) ) {
                                      perror( filename );
                                      return -1;
                                  }
                                  col += 2;
                              }
                              else {
                                  printf("Unsupported control code 0x%02x on position %d of field '%s' ('%s').\n", field->data[i], i, field->template->name, field->data);
                                  return -1;
                              }
                          }
                      }
#if 0
                      /* did we hit the max? */
                      if ( field->template->maxlen &&
                           field->len >= field->template->maxlen )
                          break;
#endif
                  }
                  if ( 0 > fprintf(file,"\n") ) {
                      perror( filename );
                      return -1;
                  }
              }
              break;

          case binary:
              {
#if 0
                  int i;
                  int col=0;
#endif

                  if ( 0 > fprintf(file, "%s:%d:\n:",
                                   field->template->name,
                                   field->len) ) {
                      perror( filename );
                      return -1;
                  }

#if 0
                  /* escape data */
                  for (i=0; i<field->len; i++) {
                      col += 2;
                      if ( col > 70 ) {
                          fprintf(file,"\n:");
                          col = 1;
                      }
                      if ( 0 > fprintf(file,"%02x", field->data[i]) ) {
                          perror( filename );
                          return -1;
                      }
                  }

                  if ( 0 > fprintf(file,"\n") ) {
                      perror( filename );
                      return -1;
                  }

#else
                  /* create a base64 encoded stream */
                  base64Encode(file, field->data, field->len);
#endif
              }
              break;

          default:
              printf("Internal error (bad type) at line %d.\n", __LINE__ );
              return -1;
        }
    }

    return 0;
}

/* read a database template file */
int readtemplate( char* database )
{
    FILE* file;
    char line[128];
    char dbname[256];

    /* initiate field list */
    InitList( &dbTemplate, NO_SIZE_CHECK);

    sprintf(dbname, "%s%s", database, DB_TEMPLATE_EXTENSION );
    file = fopen( dbname, "r" );
    if (!file)
        return -1;

    while ( fgets(line, 128, file) ) {
        int a;
        char name[80], type[16];

        if (isspace(line[0]) || ('#'==line[0]))
            continue;

        /* reasonable version */
        a=sscanf(line,
                 "%63[^:]:" /* field name */
                 "%8s",  /* field type */
                 name, type );

        /* did we get it all? */
        if ( 2 <= a ) {
            t_fieldTemplate* newfield = (t_fieldTemplate*) malloc( sizeof( t_fieldTemplate ));
            
            strcpy( newfield->name, name );
            switch ( tolower(type[0]) ) {
              case 'b': newfield->type = binary; break;
              case 's': newfield->type = string; break;
              case 'i': newfield->type = integer; break;
              default:
                  printf("Unknown field type '%s'!\n", type );
                  fclose( file );
                  return -1;
            }
            AddLast( &dbTemplate, &newfield->node );

/*            printf(""); */

/*            printf("'%s' of type %d\n", newfield->name, newfield->type ); */
            if ( newfield->name == NULL )
                return -1;

            if ( newfield->type == 0 )
                return -1;
        }
    }

    fclose( file );

    return 0;
}

/* read a database */
/* params: 1: 'read' 2:filename */
int readdb( char* database )
{
    FILE* file;
    char dbname[256],line[128];
    int a,l,linenum=0;
    t_fieldType type;
    t_row* row=NULL;
    char* dataptr;

    if (readtemplate( database ))
        return -1;

    /* create .db filename */
    sprintf( dbname, "%s%s", database, DB_FILE_EXTENSION );
    file = fopen( dbname, "r" );
    if (!file) {
        perror( dbname );
        return -1;
    }

    /* init rowlist */
    InitList( &rowList, NO_SIZE_CHECK );

    /* read lines */
    while ( fgets(line, 128, file) ) {
        char name[80], len[16];
        unsigned int value;

        linenum++;

        switch ( line[0] ) {

          case '-': { /* new row */
              int i;

              row = (t_row*) malloc( sizeof( t_row ) );
              AddLast( &rowList, &row->node );

              /* read id */
              i = sscanf( line, "- %08lx%08lx", &row->id1, &row->id2 );
              if ( i != 2 ) {
                  printf("Bad row id on line %d!\n", linenum );
                  return -1;
              }

              InitList( &row->fieldList, NO_SIZE_CHECK );

              break;
          }

          case ':': /* continued field */
              if (!row) {
                  printf("Database error: Data before new-row marker!\n");
                  return -1;
              }
              
              switch ( type ) {
              case string: {
                  int i,len,x=0;

                  /* remove trailing NL */
                  len = strlen( line );
                  if ( '\n' == line[len-1] ) {
                      line[ len - 1 ] = 0;
                      len--;
                  }

                  /* parse escaped chars */
                  for (i=1; i<len; i++) {
                      if ( '\\' == line[i] ) {
                          i++;
                          switch ( line[i] ) {
                            case 'n': dataptr[x++] = '\n'; break;
                            case 't': dataptr[x++] = '\t'; break;
                            case 'r': dataptr[x++] = '\r'; break;
                            case '\\': dataptr[x++] = '\\'; break;
                            default:
                                printf("Error: Unsupported control code '\\%c'in data!\n", line[i]);
                                return -1;
                          }
                      }
                      else
                          dataptr[x++] = line[i];
                  }

                  /* move pointer for next line */
                  dataptr += x;
                  break;
              }
                  
              case binary: {
                  int len;

                  /* remove trailing NL */
                  len = strlen( line );
                  if ( '\n' == line[len-1] ) {
                      line[ len - 1 ] = 0;
                      len--;
                  }
                  l=strlen( line+1 );
#if 0
                  for ( i=0; i<l; i+=2) {
                      int val;
                      if (!sscanf( line+1+i, "%02x", &val )) {
                          printf("Bad data in line %d of database!\n",linenum);
                          return -1;
                      }
                      *dataptr = val;
                      dataptr++;
                  }
#else
                  /* base64 decode the input: */
                  base64Decode(dataptr, &dataptr,
                               line+1, l);
#endif
                  break;
              }
                  
              default:
                  printf("Internal error (bad type:%d) when reading line %d\n",
                         type, linenum );
                  return -1;
              }
              break;
        
          default: { /* new field */
              t_field* newfield;
              t_fieldTemplate* template;
              
              if (!row) {
                  printf("Database error: Data before new-row marker!\n");
                  return -1;
              }

              newfield = (t_field*) malloc( sizeof( t_field ) );
              AddLast( &row->fieldList, &newfield->node );
              
              a=sscanf(line,
                       "%63[^:]:" /* field name */
                       "%16[^:]:" /* field length */
                       "%d",      /* field data (if integer) */
                       name, len, &value );
              
              /* find template for field */
              for ( template = (t_fieldTemplate*) GetFirst( &dbTemplate ); 
                   template;
                   template = (t_fieldTemplate*) GetNext( &template->node ) ) {

                  if ( !strcmp( name, template->name ) )
                      break;
              }

              if (!template)
                  continue;
              
              type = template->type;
              newfield->template = template;
                  
              if ( integer == type )
                  newfield->value = value;
              else {
                  newfield->len = atoi( len );
                  dataptr = (unsigned char*) malloc( newfield->len );
                  newfield->data = dataptr;
              }
          }
        }
    }

    return 0;
}

/* output data according to format string */
int print_row( t_row *row, char* format, t_textType textMode )
{
  int i;
  int flen = strlen(format);
  
  for ( i=0; i<flen; i++ ) {
    switch ( format[i] ) {

    case '\\': /* control code */
      i++;
      switch ( textMode ) {
        case url:
          switch ( format[i] ) {
            case '\\': printf("%%%02X", '\\'); break;
            case 'n': printf("%%%02X", '\n'); break;
            case 't': printf("%%%02X", '\t'); break;
            case '.': break; /* used for concatening fields */
            default: printf("%%%02X", format[i] ); break;
          }
          break;

        case html:
          switch ( format[i] ) {
            case '\\': printf("&#%d;", '\\'); break;
            case 'n': printf("&#%d;", '\n'); break;
            case 't': printf("&#%d;", '\t'); break;
            case '.': break; /* used for concatening fields */
            default: printf("&#%u;", (unsigned char)format[i] ); break;
          }
          break;

        case text:
          switch ( format[i] ) {
            case '\\': printf("\\"); break;
            case 'n': printf("\n"); break;
            case 't': printf("\t"); break;
            case '.': break; /* used for concatening fields */
            default: printf("%c", format[i] ); break;
          }
          break;
      }
      break;
                
    case '%': { /* field expansion */
      int x;
      char fieldname[32];
      t_field* field;

      i++;

      /* support % */
      if ( '%'==format[i] ) {
          switch ( textMode ) {
            case url: printf("%%%02X", '%'); break;
            case html: printf("&#%d;", '%'); break;
            case text: printf("%%"); break;
          }
	continue;
      }

      /* support # (row id) */
      if ( '#'==format[i] ) {
	printf("%08lx%08lx", row->id1, row->id2 );
	continue;
      }
                  
      /* get field name */
      for ( x=0; isalnum(format[i+x]); x++ )
	fieldname[x] = format[i+x];
      fieldname[x] = 0;
      i+=x-1;
                  
      field = getfield( row, fieldname );
      if (!field)
	break;

      /* output field */
      switch ( field->template->type ) {
      case integer:
	printf("%d", field->value );
	break;
                      
      case string:
          switch ( textMode ) {
            case url: {
              char* tmp = string2url( field->data );
              printf("%s", tmp );
              free( tmp );
              break;
            }

            case html: {
              char* tmp = string2html( field->data );
              printf("%s", tmp );
              free( tmp );
              break;
            }

            case text:
              printf("%s", field->data );
              break;
          }
          break;

      case binary:
	fwrite( field->data, field->len, 1, stdout );
	break;

      default:
	printf("Internal error (wrong type:%d)\n",
	       field->template->type);
	return -1;
      }
      break;
    }

    default:
      /* print char */
      switch ( textMode ) {
        case text:
          printf("%c", format[i]);
          break;

        case url:
          if ( isalnum(format[i]))
              printf("%c", format[i]);
          else
              if ( ' ' == format[i] )
                  printf("+");
              else
                  printf("%%%02X", format[i]);
          break;

        case html:
          if ( isalnum(format[i]))
              printf("%c", format[i]);
          else
              printf("&#%u;", (unsigned char)format[i]);
          break;
      }
      break;
    }
  }

  return 0;
}

