
/* 
   Author:  Daniel Stenberg <Daniel.Stenberg@sth.frontec.se>
   Version: 0.1

   This is a Base64 encoder/decoder as defined in RFC 2045.

   */

#include <stdio.h>
#include <string.h>

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

static char table64[]=
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  
void base64Encode(FILE *file, char *intext, int length)
{
  unsigned char ibuf[3];
  unsigned char obuf[4];
  int i;
  int inputparts;
  int column=1;

  while(length) {
    for (i = inputparts = 0; i < 3; i++) { 
      if(length) {
        inputparts++;
        ibuf[i] = *intext;
        intext++;
        length--;
      }
      else
        ibuf[i] = 0;
    }
    
    obuf [0] = (ibuf [0] & 0xFC) >> 2;
    obuf [1] = ((ibuf [0] & 0x03) << 4) | ((ibuf [1] & 0xF0) >> 4);
    obuf [2] = ((ibuf [1] & 0x0F) << 2) | ((ibuf [2] & 0xC0) >> 6);
    obuf [3] = ibuf [2] & 0x3F;

    switch(inputparts) {
    case 1: /* only one byte read */
      fprintf(file, "%c%c==", 
             table64[obuf[0]],
             table64[obuf[1]]);
      break;
    case 2: /* two bytes read */
      fprintf(file, "%c%c%c=", 
             table64[obuf[0]],
             table64[obuf[1]],
             table64[obuf[2]]);
      break;
    default:
      fprintf(file, "%c%c%c%c", 
             table64[obuf[0]],
             table64[obuf[1]],
             table64[obuf[2]],
             table64[obuf[3]] );
      break;
    }
    column += 4;
    if ( column > 70 ) {
      fprintf(file, "\n:");
      column = 1;
    }
  }
  fprintf(file, "\n");
}


void base64Decode(char *outtext, char **outend,
                  char *intext, int length)
{
  unsigned char ibuf[4];
  unsigned char obuf[3];
  char ignore;
  char endtext=FALSE;
  char ch;
  int index=0;

  while(length--) {
    ch = *intext;
    ignore = FALSE;
    if ((ch >= 'A') && (ch <= 'Z'))
      ch = ch - 'A';
    else if ((ch >= 'a') && (ch <= 'z'))
      ch = ch - 'a' + 26;
    else if ((ch >= '0') && (ch <= '9'))
      ch = ch - '0' + 52;
    else if (ch == '+')
      ch = 62;
    else if (ch == '=') /* end of text */
      endtext = TRUE;
    else if (ch == '/')
      ch = 63;
    else
      ignore = TRUE; 
    
    if (!ignore) {
      if(!endtext) {
        ibuf[index] = ch;

        index ++;
        index &= 3; /* use bit arithmetic instead of remainder */
      }
      if ((0 == index) || endtext) {

        obuf [0] = (ibuf [0] << 2) | ((ibuf [1] & 0x30) >> 4);
        obuf [1] = ((ibuf [1] & 0x0F) << 4) | ((ibuf [2] & 0x3C) >> 2);
        obuf [2] = ((ibuf [2] & 0x03) << 6) | (ibuf [3] & 0x3F);

        switch(index) {
        case 1:
          sprintf(outtext, "%c", obuf[0]);
          break;
        case 2:
          sprintf(outtext, "%c%c", obuf[0], obuf[1]);
          break;
        default:
          sprintf(outtext, "%c%c%c", obuf[0], obuf[1], obuf[2]);
          break;
        }
        outtext += strlen(outtext);
        memset(ibuf, 0, sizeof(ibuf));
      }      
    }
    intext++;
  }
  *outend = outtext;
}
