/*************************************************************************
 *
 *  Frexx Database
 *
 *  A simple, lightweight database 
 *
 ************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <memory.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "types.h"
#include "lists.h"
#include "dbase.h"

#include "expr.h"

BOOL Match(char *string, char *pattern, BOOL case_insensitive);

/*
** --- global variables ---
*/
extern LIST dbTemplate;
extern LIST rowList;

EXPR_RESULT MatchFunc(EXPR_NODE *def, EXPR_NODE *res, void *data)
{
  EXPR_RESULT err;
  EXPR_NODE par1;
  EXPR_NODE par2;

  err = CalculateTree(def->par[0], &par1, data);
  if(err)
    return(err);
  if(!IS_STRING(par1.type))
    return EXPR_INVALID_TYPE;
  
  err = CalculateTree(def->par[1], &par2, data);
  if(err)
    return(err);
  if(!IS_STRING(par2.type))
    return EXPR_INVALID_TYPE;

  res->num = Match(par1.str, par2.str, FALSE);
  res->type = COMP_INTEGER;
  debugf("Function result: %d\n", res->num);
  return EXPR_OK;
}

EXPR_RESULT IMatchFunc(EXPR_NODE *def, EXPR_NODE *res, void *data)
{
  EXPR_RESULT err;
  EXPR_NODE par1;
  EXPR_NODE par2;

  err = CalculateTree(def->par[0], &par1, data);
  if(err)
    return(err);
  if(!IS_STRING(par1.type))
    return EXPR_INVALID_TYPE;
  
  err = CalculateTree(def->par[1], &par2, data);
  if(err)
    return(err);
  if(!IS_STRING(par2.type))
    return EXPR_INVALID_TYPE;

  res->num = Match(par1.str, par2.str, TRUE);
  res->type = COMP_INTEGER;
  debugf("Function result: %d\n", res->num);
  return EXPR_OK;
}

int find( int argc, char* argv[] )
{
  EXPR_RESULT error;
  EXPR_FUNCTION functions[] =
  {
    {"match", MatchFunc, COMP_INTEGER, 2, 0},
    {"imatch", IMatchFunc, COMP_INTEGER, 2, 0},
    {NULL, 0, 0}
  };
  EXPR_OPER_LIST exprList;
  EXPR_NODE *tree;
  EXPR_NODE res;
  t_row *row;
  char* format = argv[4];
  t_textType textMode=text;

  if ( argc < 5 ) {
    printf("usage: %s find [database] [expression] {-u | -h} [format]\n", argv[0]);
    return -1;
  }
  
  if ( readdb( argv[2] ) ) {
    perror( argv[2] );
    return -1;
  }

  if ( '-' == argv[4][0] ) {
      format = argv[5];
      switch ( argv[4][1] ) {
        case 'u': textMode=url; break;
        case 'h': textMode=html; break;
        default: break;
      }
  }

  InitList(&exprList.list, NO_SIZE_CHECK);
  error = ParseExpression(argv[3], NULL, NULL, functions, &exprList);
  debugf("Number of nodes in parsed list: %li\n", exprList.list.num_nodes);
  if(EXPR_OK != error)
  {
    printf("Expression error -  %s\n", ExprErrorText(error));
    DeleteExpression(&exprList);
    return -1;
  }
  else
  {
    tree = CreateTree(&exprList, FALSE);
#ifdef DEBUG
    OutputTree(tree);
#endif

    for ( row = (t_row*) GetFirst( &rowList ); row;
          row = (t_row*) GetNext( &row->node ) )
    {
      error = CalculateTree(tree, &res, row);
      if(EXPR_OK != error)
      {
	printf("Expression error - %s\n", ExprErrorText(error));
	DeleteExpression(&exprList);
	return -1;
      }

      if(res.num)
      {
	if(print_row(row, format, textMode ))
	{
	  DeleteExpression(&exprList);
	  return -1;
	}
      }
    }
    DeleteExpression(&exprList);
  }
  return 0;
}

/*
  The shortest ever? 88 characters...
  m(char*s,char*t){return*t-'*'?*s?*t=='?'|*s==*t&&m(s+1,t+1):!*t:m(s,t+1)||*s&&m(s+1,t);}

  Written by Staffan Ulfberg <staffanu@tcs07.nada.kth.se>

  What follows below is a rewrite of the above into readable C and a small
  patch to prevent really deep recursiveness on patterns featuring a long
  sequence of '*'. Done by me. Daniel Stenberg
  <Daniel.Stenberg@sth.frontec.se>
 */
#include "types.h"
#include <ctype.h>

BOOL Match(char *string, char *pattern, BOOL case_insensitive)
{
  for (; '*'^*pattern; ++pattern, ++string) {
    if (!*string)
      return (!*pattern);
    if(case_insensitive) {
      if (toupper(*string)^toupper(*pattern) && '?'^*pattern)
	return FALSE;
    }
    else {
      if (*string^*pattern && '?'^*pattern)
	return FALSE;
    }    
  }
  /* two-line patch to prevent *too* much recursiveness: */
  while('*' == pattern[1])
    pattern++;
  do {
    if ( Match(string, pattern + 1, case_insensitive) )
      return TRUE;
  } while (*string++);
  return FALSE;
}
